package ru.dbotthepony.kstarbound.util

import ru.dbotthepony.kommons.util.MailboxExecutorService
import ru.dbotthepony.kstarbound.Starbound
import java.util.concurrent.locks.LockSupport
import java.util.function.BooleanSupplier

class ExecutionSpinner(private val executor: MailboxExecutorService, private val spinner: BooleanSupplier, private val timeBetweenFrames: Long, val precise: Boolean = false) : Runnable {
	private var lastRender = System.nanoTime()
	private var frameRenderTime = 0L
	private val frameRenderTimes = LongArray(60) { 1L }
	private var frameRenderIndex = 0
	private val renderWaitTimes = LongArray(60) { 1L }
	private var renderWaitIndex = 0

	val averageRenderWait: Double get() {
		var sum = 0.0

		for (value in renderWaitTimes)
			sum += value

		if (sum == 0.0)
			return 0.0

		sum /= 1_000_000_000.0
		return sum / renderWaitTimes.size
	}

	val averageRenderTime: Double get() {
		var sum = 0.0

		for (value in frameRenderTimes)
			sum += value

		if (sum == 0.0)
			return 0.0

		sum /= 1_000_000_000.0
		return sum / frameRenderTimes.size
	}

	private fun timeUntilNextFrame(): Long {
		return Starbound.TICK_TIME_ADVANCE_NANOS - (System.nanoTime() - lastRender) - frameRenderTime
	}

	fun spin(): Boolean {
		var diff = timeUntilNextFrame()

		if (precise) {
			while (diff > 1_500_000L) {
				executor.executeQueuedTasks()
				diff = timeUntilNextFrame()
				if (diff > 1_500_000L) LockSupport.parkNanos(diff - 400_000L)
			}

			while (diff > 0L) {
				executor.executeQueuedTasks()
				diff = timeUntilNextFrame()
			}
		} else {
			while (diff > 0L) {
				executor.executeQueuedTasks()
				diff = timeUntilNextFrame()
				if (diff > 400_000L) LockSupport.parkNanos(diff)
			}
		}

		val mark = System.nanoTime()
		val result = spinner.asBoolean
		frameRenderTime = System.nanoTime() - mark
		frameRenderTimes[++frameRenderIndex % frameRenderTimes.size] = frameRenderTime
		renderWaitTimes[++renderWaitIndex % renderWaitTimes.size] = System.nanoTime() - lastRender
		lastRender = System.nanoTime()

		return result
	}

	override fun run() {
		while (spin()) {}
	}
}
